import click
import json
import os


_common_help = 'Some kind of help for molecules'


_short_help = 'Perform operations on molecules'


@click.group('molecules', short_help=_short_help, help='%s\n\n%s' % (
    _short_help,
    _common_help))
@click.pass_obj
def molecules(gc):
    pass


@molecules.command('ls')
@click.option('--limit', default=25, help='The results limit')
@click.pass_obj
def _list(gc, limit):
    params = {'limit': limit}
    molecules = gc.get('/molecules', parameters=params)
    print(_format_molecules(molecules))


@molecules.command('upload')
@click.argument('filename')
@click.option('--format', default=None, help='The format of the file')
@click.pass_obj
def _upload(gc, filename, format):

    if format is None:
        format = os.path.splitext(filename)[1][1:].lower()

    with open(filename, 'r') as rf:
        contents = rf.read()

    body = {
      format: contents
    }

    mol = gc.post('/molecules', json=body)
    print('Molecule uploaded:')
    print(_format_molecule(mol))


def _format_molecules(molecules):
    ret = 'matches: ' + str(molecules['matches']) + '\n'
    ret += 'limit: ' + str(molecules['limit']) + '\n'
    ret += 'offset: ' + str(molecules['offset']) + '\n'
    ret += 'results:\n'
    for res in molecules['results']:
        ret += _format_molecule(res) + '\n'

    return ret


def _format_molecule(mol):
    return json.dumps(mol, indent=4)



import json
import os
import re

from girder_client import HttpError

from . import avogadro


def sif_dir():
    return '$HOME/.oc/singularity'


def digest_to_sif(digest):
    return os.path.join(sif_dir(), digest + '.sif')


def image_to_sif(image_str):
    # Let's remove special characters from the image string
    name = '%s.sif' % re.sub('[^a-zA-Z0-9]', '_', image_str)
    return os.path.join(sif_dir(), name)


def cjson_to_xyz(cjson):
    xyz = avogadro.convert_str(json.dumps(cjson), 'cjson', 'xyz')
    return xyz


def get_cori(client):
    params = {
        'type': 'newt'
    }
    clusters = client.get('clusters', parameters=params)
    for cluster in clusters:
        if cluster['name'] == 'cori':
            return cluster

    # We need to create one
    body = {
        'config': {
            'host': 'cori'
        },
        'name': 'cori',
        'type': 'newt'
    }
    cluster = client.post('clusters', data=json.dumps(body))

    return cluster


def log_and_raise(task, msg):
    log_error(task, msg)
    raise Exception(msg)


def log_error(task, msg):
    task.taskflow.logger.error(msg)


def log_std_err(task, client, run_folder):
    errors = get_std_err(client, run_folder)
    for e in errors:
        log_error(task, e)


def get_std_err(client, run_folder):
    error_regex = re.compile(r'^.*\.e\d*$', re.IGNORECASE)
    output_items = list(client.listItem(run_folder['_id']))
    errors = []
    for item in output_items:
        if error_regex.match(item['name']):
            files = list(client.listFile(item['_id']))
            if len(files) != 1:
                continue
            with tempfile.TemporaryFile() as tf:
                client.downloadFile(files[0]['_id'], tf)
                tf.seek(0)
                contents = tf.read().decode()
                errors.append(contents)
    return errors


def get_oc_folder(client):
    me = client.get('user/me')
    if me is None:
        raise Exception('Unable to get me.')

    login = me['login']
    private_folder_path = 'user/%s/Private' % login
    private_folder = client.resourceLookup(private_folder_path)
    oc_folder_path = '%s/oc' % private_folder_path
    # girder_client.resourceLookup(...) no longer has a test parameter
    # so we just assume that if resourceLookup(...) raises a HttpError
    # then the resource doesn't exist.
    try:
        oc_folder = client.resourceLookup(oc_folder_path)
    except HttpError:
        oc_folder = client.createFolder(private_folder['_id'], 'oc')

    return oc_folder


def is_nersc(cluster):
    return cluster.get('name') in ['cori']


def is_demo(cluster):
    return cluster.get('name') == 'demo_cluster'


def countdown(cluster):
    """
    Returns the number of seconds the monitoring task should be delayed before
    running based on the cluster we are using.
    """
    countdown = 0
    # If we are running at NERSC our job states are cached for 60 seconds,
    # so they are potentially 60 seconds out of date, so we have to wait
    # at least 60 seconds before we can assume that the job is complete if its
    # nolonger in the queue, so we delay our monitoring
    if is_nersc(cluster):
        countdown = 65

    return countdown


def post_image_to_database(client, container, repository, tag, digest, cluster,
                           size):
    body = {
        'type': container,
        'repository': repository,
        'tag': tag,
        'digest': digest,
        'clusterId': cluster['_id'],
        'size': size
    }

    try:
        client.post('images', body)
    except HttpError as e:
        # Just ignore the error if the image already exists
        if e.status != 409:
            raise



import os

from ._singleton import Singleton
from ._utils import get_oc_token_obj

@Singleton
class Application(object):

    def __init__(self):
        token_obj = get_oc_token_obj()
        url = token_obj.get('appUrl')
        url = os.environ.get('OC_APP_URL', url)
        self.url = url



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

install_reqs = [
    'girder_client>=2.3.0',
    'click'
]

setup(
    name='openchemistry_client',
    setup_requires=['setuptools_scm'],
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/OpenChemistry/openchemistrypy',
    author='Kitware Inc',
    license='BSD 3-Clause',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='',
    packages=find_packages(),
    install_requires=install_reqs,
    entry_points={
        'console_scripts': [
            'occlient = openchemistry_client.client:main'
        ]
    }
)



import json
import argparse
import subprocess
from subprocess import PIPE
import os
import errno


def split_image(img):
    split = img.split(':')
    repository = split[0]
    tag = split[1] if len(split) > 1 else 'latest'
    return repository, tag


def run_subprocess(args):
    # Runs subprocess and returns a tuple of the output and the error
    p = subprocess.Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, err = p.communicate()
    rc = p.returncode
    if rc != 0:
        raise Exception('Command failed: ' + args.join(' '))

    return output.decode('utf-8'), err.decode('utf-8')


def ensure_singularity_dir():
    home = os.path.expanduser("~")
    singularity_dir = os.path.join(home, '.oc', 'singularity')

    try:
        os.makedirs(singularity_dir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(singularity_dir):
            pass
        else:
            raise

    return singularity_dir


def singularity():
    singularity_dir = ensure_singularity_dir()

    images = []
    for f in os.listdir(singularity_dir):
        if not f.endswith('.sif'):
            continue

        # The digest is the name of the file, without .sif
        digest = f.replace('.sif', '')

        path = os.path.join(singularity_dir, f)
        output, err = run_subprocess(['singularity', 'inspect', path])

        info = json.loads(output)
        repo_tag_str = 'org.label-schema.usage.singularity.deffile.from'
        if repo_tag_str not in info:
            continue

        repository, tag = split_image(info[repo_tag_str])

        size = os.path.getsize(path)

        image = {
            'repository': repository,
            'tag': tag,
            'digest': digest,
            'size': size
        }
        images.append(image)

    return images


def docker():
    ls_args = ['docker', 'image', 'ls']
    output, err = run_subprocess(ls_args)

    images = []
    lines = output.split('\n')
    if lines:
        # Remove the header line
        lines.pop(0)

    for line in lines:
        line_split = line.split()
        if len(line_split) < 5:
            continue

        image_id = line_split[2]

        inspect_args = ['docker', 'inspect', image_id]
        output, err = run_subprocess(inspect_args)

        info = json.loads(output)
        if not info:
            continue

        info = info[0]

        if not info.get('RepoTags'):
            continue

        repository, tag = split_image(info['RepoTags'][0])

        if not info.get('RepoDigests'):
            # This will be empty for user-created images
            continue

        _, digest = split_image(info['RepoDigests'][0])

        size = info['Size']
        image = {
            'repository': repository,
            'tag': tag,
            'digest': digest,
            'size': size
        }
        images.append(image)

    return images


def shifter():
    output, err = run_subprocess(['shifterimg', 'images'])

    # FIXME: the below is just a guess, and probably doesn't work
    # I don't have access to NERSC to test this.
    images = []
    for line in output.split('\n'):
        line_split = line.split()
        if len(line_split) < 6:
            continue

        repository, tag = split_image(line_split[5])

        # Get the full digest
        digest, err = run_subprocess(['shifterimg', 'lookup', line_split[5]])

        # FIXME: implement a way to get the shifter image size
        # I am currently not aware of one, and don't have access to NERSC
        size = 0

        image = {
            'repository': repository,
            'tag': tag,
            'digest': digest,
            'size': size
        }
        images.append(image)

    return images


def write_descriptor(images):
    with open('list.json', 'w') as fp:
        json.dump(images, fp)


def main():
    parser = argparse.ArgumentParser(
        description='Pull image from container registry.')
    parser.add_argument('-c', '--container',
                        choices=['docker', 'singularity', 'shifter'],
                        help='the type of container', required=True)

    args = parser.parse_args()
    container = args.container

    if container == 'singularity':
        images = singularity()
    elif container == 'shifter':
        images = shifter()
    else:
        images = docker()

    write_descriptor(images)


if __name__ == "__main__":
    main()



class Singleton(object):
    """Singleton decorator
    """
    def __init__(self, cls):
        self._cls = cls
        self._instance = self._cls()

    def __call__(self):
        return self._instance



import click
import json

import openchemistry as oc
from openchemistry._calculation import CalculationResult
from openchemistry._calculation import GirderMolecule
from openchemistry._calculation import PendingCalculationResultWrapper


_common_help = 'Some kind of help for calculations'


_short_help = 'Perform operations on calculations'


@click.group('calculations', short_help=_short_help, help='%s\n\n%s' % (
    _short_help,
    _common_help))
@click.pass_obj
def calculations(gc):
    pass


@calculations.command('ls')
@click.option('--limit', default=25, help='The results limit')
@click.pass_obj
def _list(gc, limit):
    params = {'limit': limit}
    calcs = gc.get('/calculations', parameters=params)
    print(_format_calculations(calcs))


@calculations.command('submit')
@click.argument('input_file')
@click.pass_obj
def _submit(gc, input_file):

    # Make sure openchemistry is using our girder client...
    _patch_oc_girder_client(gc)

    with open(input_file, 'r') as rf:
        input_dict = json.load(rf)

    molecules = [GirderMolecule(input_dict['molecule_id'])]
    image_name = input_dict['image_name']
    input_parameters = input_dict['input_parameters']
    result = oc.run_calculations(molecules, image_name, input_parameters)[0]

    if isinstance(result, PendingCalculationResultWrapper):
        print('Calculation running:', result.unwrap()._id)
    elif isinstance(result, CalculationResult):
        print('Calculation already performed:', result._id)
    else:
        print('Unknown type returned from oc.run_calculations():',
              type(result))


@calculations.command('download')
@click.argument('calculation_id')
@click.option('-o', '--output-file', default='oc_calculation.json',
              help='The file to place the output cjson into.')
@click.pass_obj
def _download(gc, calculation_id, output_file):
    calcs = gc.get('/calculations/%s' % calculation_id)
    if 'properties' in calcs and calcs['properties'].get('pending', False):
        print('Calculation is currently running')
        return

    print('Downloading calculation:', calculation_id)
    cjson = calcs.get('cjson', {})

    print('Writing cjson to output file:', output_file)
    with open(output_file, 'w') as wf:
        json.dump(cjson, wf)


def _format_calculations(calcs):
    ret = 'matches: ' + str(calcs['matches']) + '\n'
    ret += 'limit: ' + str(calcs['limit']) + '\n'
    ret += 'offset: ' + str(calcs['offset']) + '\n'
    ret += 'results:\n'
    for res in calcs['results']:
        ret += _format_calculation(res) + '\n'

    return ret


def _format_calculation(calc):
    return json.dumps(calc, indent=4)


def _patch_oc_girder_client(gc):
    # Make sure the OC girder client is using our client...
    oc._girder.GirderClient()
    setattr(oc._girder.GirderClient.__dict__['instance'], 'client', gc)



import cumulus
from cumulus.taskflow import TaskFlow
from cumulus.taskflow import logging
from cumulus.taskflow.cluster import create_girder_client
from cumulus.tasks.job import monitor_job, submit_job
from cumulus.tasks.job import (
    download_job_input_folders, upload_job_output_to_folder
)

from girder_client import HttpError
from girder.api.rest import getCurrentUser
from girder.constants import AccessType
from girder.utility.model_importer import ModelImporter

from .utils import (
    get_cori, get_oc_folder, log_and_raise, is_demo, is_nersc, countdown,
    post_image_to_database
)

import datetime
import io
import json
from jsonpath_rw import parse
import os
import re

STATUS_LEVEL = logging.INFO + 5
PYTHON_FILE = 'list.py'
OUTPUT_FILE = 'list.json'

# Only these repositories will be registered automatically
REPOSITORIES_TO_REGISTER = [
    'openchemistry/chemml',
    'openchemistry/cp2k',
    'openchemistry/nwchem',
    'openchemistry/psi4',
    'openchemistry/torchani'
]

class ContainerListTaskFlow(TaskFlow):
    """
    {
        "cluster": {
            "_id": <id of cluster to run on>
        },
        'container': <container technology to be used: docker | singularity | shifter>
    }
    """

    def start(self, *args, **kwargs):
        user = getCurrentUser()
        cluster = kwargs.get('cluster')
        container = kwargs.get('container')

        if cluster is None:
            raise Exception('Unable to extract cluster.')

        if '_id' not in cluster and 'name' not in cluster:
            raise Exception('Unable to extract cluster.')

        if container is None:
            raise Exception('Unable to extract container type.')

        cluster_id = parse('cluster._id').find(kwargs)
        if cluster_id:
            cluster_id = cluster_id[0].value
            model = ModelImporter.model('cluster', 'cumulus')
            cluster = model.load(cluster_id, user=user, level=AccessType.ADMIN)

        super(ContainerListTaskFlow, self).start(
            start.s(user, cluster, container),
            *args, **kwargs)


def _get_job_parameters(task, cluster, container):
    setup_commands = []
    job_parameters = {
        'taskFlowId': task.taskflow.id
    }

    # Override default parameters depending on the cluster we are running on
    if is_nersc(cluster):
        # NERSC specific options
        job_parameters.update({
            'numberOfNodes': 1,
            'queue': 'debug',
            'constraint': 'haswell',
            'account': os.environ.get('OC_ACCOUNT')
        })
    elif is_demo(cluster):
        # DEV/DEMO environment options
        setup_commands = ['source scl_source enable python27']
    else:
        # baremetal options
        pass

    return {
        'container': container,
        'setupCommands': setup_commands,
        'jobParameters': job_parameters
    }


def _create_job(task, cluster, folder, container):
    params = _get_job_parameters(task, cluster, container)
    setup_commands = params['setupCommands']
    job_parameters = params['jobParameters']

    run_command = 'python %s -c %s' % (PYTHON_FILE, container)
    commands = setup_commands + [
        run_command,
        'rm %s' % PYTHON_FILE
    ]

    body = {
        # ensure there are no special characters in the submission script name
        'name': 'desc_%s' % re.sub('[^a-zA-Z0-9]', '_', run_command),
        'commands': commands,
        'input': [
            {
              'folderId': folder['_id'],
              'path': '.'
            }
        ],
        'output': [
            {
              'folderId': folder['_id'],
              'path': '.'
            }
        ],
        'uploadOutput': False,
        'params': job_parameters
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))

    return job


@cumulus.taskflow.task
def start(task, user, cluster, container):
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    if cluster.get('name') == 'cori':
        cluster = get_cori(client)

    if '_id' not in cluster:
        log_and_raise(task, 'Invalid cluster configurations: %s' % cluster)

    oc_folder = get_oc_folder(client)
    root_folder = client.createFolder(oc_folder['_id'],
                                      datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%f"))
    # temporary folder to save the container in/out
    folder = client.createFolder(root_folder['_id'], 'list_folder')

    # save the list.py script to the job directory
    with open(os.path.join(os.path.dirname(__file__), 'utils/%s' % PYTHON_FILE), 'rb') as f:
        # Get the size of the file
        size = f.seek(0, 2)
        f.seek(0)
        client.uploadFile(folder['_id'], f, PYTHON_FILE, size, parentType='folder')

    job = _create_job(task, cluster, folder, container)

    # Now download list.py script to the cluster
    task.taskflow.logger.info('Preparing job to list the images.')
    download_job_input_folders(cluster, job,
                               girder_token=task.taskflow.girder_token,
                               submit=False)

    task.taskflow.logger.info('Submitting job to list the images.')
    submit_job(cluster, job, girder_token=task.taskflow.girder_token,
               monitor=False)

    monitor_job.apply_async(
        (cluster, job), {'girder_token': task.taskflow.girder_token,
                         'monitor_interval': 10},
        countdown=countdown(cluster),
        link=postprocess_job.s(user, cluster, job, folder, container))


@cumulus.taskflow.task
def postprocess_job(task, _, user, cluster, job, folder, container):
    task.taskflow.logger.info('Finished listing the images')

    task.taskflow.logger.info('Processing the results of the list')
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    # Refresh state of job
    job = client.get('jobs/%s' % job['_id'])

    upload_job_output_to_folder(cluster, job, girder_token=task.taskflow.girder_token)

    output_items = list(client.listItem(folder['_id']))
    output_file = None
    for item in output_items:
        if item['name'] == OUTPUT_FILE:
            files = list(client.listFile(item['_id']))
            if len(files) != 1:
                log_and_raise(task, 'Expecting a single file under item, found: %s' % len(files))
            output_file = files[0]
            break

    if output_file is None:
        log_and_raise(task, 'Could not locate output file: %s' % OUTPUT_FILE)

    # Download the output file
    output_io = io.BytesIO()
    client.downloadFile(output_file['_id'], output_io)

    # Remove the folder
    client.delete('folder/%s' % folder['_id'])

    list_json = json.loads(output_io.getvalue().decode('utf-8'))

    for image in list_json:
        repository = image.get('repository')
        if repository not in REPOSITORIES_TO_REGISTER:
            continue

        tag = image.get('tag')
        digest = image.get('digest')
        # Convert size to GB
        size = round(image.get('size', 0) / 1.e9, 2)

        post_image_to_database(client, container, repository, tag, digest,
                               cluster, size)

    task.taskflow.logger.info('Success!')



import json
import os

from .base import BaseReader
from .constants import EV_TO_J_MOL
from .utils import _cleanup_cclib_cjson

class Cp2kReader(BaseReader):

    def read(self):
        import cclib

        cjson = {}

        project_name = None
        energy = None

        with open(self._file, 'r') as rf:
            for line in rf:
                line = line.strip()

                if line.startswith('GLOBAL| Project name'):
                    project_name = line.split()[-1]

                if line.startswith('ENERGY|'):
                    # This might occur multiple times in the file.
                    # Each will be over-written by the last.
                    energy = float(line.split()[-1]) * EV_TO_J_MOL

        if project_name:
            dir_name = os.path.dirname(self._file)
            geometry_file = dir_name + '/' + project_name + '-pos-1.xyz'
            if os.path.exists(geometry_file):
                data = cclib.io.ccread(geometry_file)
                cjson = json.loads(cclib.ccwrite(data, outputtype='cjson',))
                # Cleanup original cjson
                _cleanup_cclib_cjson(cjson)

        if energy:
            cjson.setdefault('properties', {})['totalEnergy'] = energy

        return cjson



import os
from girder_client import GirderClient as GC

from ._singleton import Singleton
from ._utils import get_oc_token_obj

@Singleton
class GirderClient(object):
    def __init__(self):
        token_obj = get_oc_token_obj()
        url = token_obj.get('apiUrl')
        api_key = token_obj.get('apiKey')

        url = os.environ.get('OC_API_URL', url)
        internal_url = os.environ.get('OC_INTERNAL_API_URL', url)
        api_key = os.environ.get('OC_API_KEY', api_key)
        token = os.environ.get('GIRDER_TOKEN')

        self.client = None
        self.url = url
        self.internal_url = internal_url

        if internal_url is not None:
            self.client = GC(apiUrl=internal_url)

            if api_key is not None:
                self.client.authenticate(apiKey=api_key)
            elif token is not None:
                self.client.token = token

    def __getattr__(self, name):
        return getattr(self.client, name)



import click
from girder_client import GirderClient

from . import calculations
from . import molecules

VERSION = '0.0.1'


@click.group()
@click.option('--api-url', default=None,
              help='RESTful API URL '
                   '(e.g https://girder.example.com:443/%s)' %
                   GirderClient.DEFAULT_API_ROOT)
@click.option('--api-key', envvar='OPENCHEMISTRY_API_KEY', default=None,
              help='[default: OPENCHEMISTRY_API_KEY env. variable]')
@click.version_option(version=VERSION,
                      prog_name='Openchemistry command line interface')
@click.pass_context
def main(ctx, api_key, api_url):
    """Openchemistry Client

    The client can be used to fetch molecules, add molecules, etc.
    """
    gc = GirderClient(apiUrl=api_url)

    if api_key is not None:
        gc.authenticate(apiKey=api_key)

    ctx.obj = gc


main.add_command(molecules.molecules)
main.add_command(calculations.calculations)


if __name__ == '__main__':
    main()



import cumulus
from cumulus.taskflow import TaskFlow
from cumulus.taskflow import logging
from cumulus.taskflow.cluster import create_girder_client
from cumulus.tasks.job import (download_job_input_folders,
                               upload_job_output_to_folder)
from cumulus.tasks.job import submit_job, monitor_job

from girder.api.rest import getCurrentUser
from girder.constants import AccessType
from girder.utility.model_importer import ModelImporter

from .utils import (
    digest_to_sif, get_cori, get_oc_folder, log_and_raise, log_std_err,
    is_demo, is_nersc, countdown
)

from jsonpath_rw import parse
import os
import datetime
import json
import tempfile
import re

STATUS_LEVEL = logging.INFO + 5

class OpenChemistryTaskFlow(TaskFlow):
    """
    {
        "input": {
            "calculations": [
                <id of pending calculation #1>
                <id of pending calculation #2>
                ...
            ],
        },
        "cluster": {
            "_id": <id of cluster to run on>
        },
        "image": {
            'repository': <the image repository, e.g. "openchemistry/psi4">
            'tag': <the image tag, e.g. "latest">
        },
        "runParameters": {
            'container': <the container technology to be used: docker | singularity | shifter>,
            'keepScratch': <whether to save the raw output of the calculations: default False>
        }
    }
    """

    def start(self, *args, **kwargs):
        user = getCurrentUser()
        input_ = kwargs.get('input')
        cluster = kwargs.get('cluster')
        image = kwargs.get('image')
        run_parameters = kwargs.get('runParameters')

        if input_ is None:
            raise Exception('Unable to extract input.')

        if '_id' not in cluster and 'name' not in cluster:
            raise Exception('Unable to extract cluster.')

        if image is None:
            raise Exception('Unable to extract the docker image name.')

        if run_parameters is None:
            run_parameters = {}

        cluster_id = parse('cluster._id').find(kwargs)
        if cluster_id:
            cluster_id = cluster_id[0].value
            model = ModelImporter.model('cluster', 'cumulus')
            cluster = model.load(cluster_id, user=user, level=AccessType.ADMIN)
            cluster = model.filter(cluster, user, passphrase=False)

        super(OpenChemistryTaskFlow, self).start(
            start.s(input_, user, cluster, image, run_parameters),
            *args, **kwargs)

@cumulus.taskflow.task
def start(task, input_, user, cluster, image, run_parameters):
    """
    The flow is the following:
    - Dry run the container with the -d flag to obtain a description of the input/output formats
    - Convert the cjson input geometry to conform to the container's expected format
    - Run the container
    - Convert the container output format into cjson
    - Ingest the output in the database
    """
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    if cluster.get('name') == 'cori':
        cluster = get_cori(client)

    if '_id' not in cluster:
        log_and_raise(task, 'Invalid cluster configurations: %s' % cluster)

    oc_folder = get_oc_folder(client)
    root_folder = client.createFolder(oc_folder['_id'],
                                    datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%f"))
    # temporary folder to save the container in/out description
    description_folder = client.createFolder(root_folder['_id'],
                                    'description')

    job = _create_description_job(task, cluster, description_folder, image, run_parameters)

    task.taskflow.logger.info('Preparing job to obtain the container description.')
    download_job_input_folders(cluster, job,
                               girder_token=task.taskflow.girder_token, submit=False)
    task.taskflow.logger.info('Submitting job to obtain the container description.')

    submit_job(cluster, job, girder_token=task.taskflow.girder_token, monitor=False)

    monitor_job.apply_async((cluster, job), {'girder_token': task.taskflow.girder_token,
                                             'monitor_interval': 10},
                                             countdown=countdown(cluster),
                            link=postprocess_description.s(input_, user, cluster, image, run_parameters, root_folder, job, description_folder))

def _get_job_parameters(task, cluster, image, run_parameters):
    container = run_parameters.get('container', 'docker') # docker | singularity
    repository = image.get('repository')
    tag = image.get('tag')
    digest = image.get('digest')
    host_dir = '$(pwd)' # the host directory being mounted into the container
    guest_dir = '/data' # the directory inside the container pointing to host_dir
    job_dir = '' # relative path from guest_dir to the job directory
    setup_commands = []
    job_parameters = {
        'taskFlowId': task.taskflow.id
    }

    if digest is None:
        digest = _get_digest(task, repository, tag)

    # Override default parameters depending on the cluster we are running on

    if is_nersc(cluster):
        # NERSC specific options
        container = 'shifter' # no root access, only singularity is supported
        job_parameters.update({
            'numberOfNodes': 1,
            'queue': 'debug',
            'constraint': 'haswell',
            'account': os.environ.get('OC_ACCOUNT')
        })
    elif is_demo(cluster):
        # DEV/DEMO environment options
        host_dir = 'dev_job_data'
        job_dir = '{{job._id}}'
        setup_commands = ['source scl_source enable python27']
    else:
        # baremetal options
        pass

    return {
        'container': container,
        'repository': repository,
        'tag': tag,
        'digest': digest,
        'hostDir': host_dir,
        'guestDir': guest_dir,
        'jobDir': job_dir,
        'setupCommands': setup_commands,
        'jobParameters': job_parameters
    }


def _create_description_job(task, cluster, description_folder, image, run_parameters):
    params = _get_job_parameters(task, cluster, image, run_parameters)
    container = params['container']
    setup_commands = params['setupCommands']
    repository = params['repository']
    tag = params['tag']
    digest = params['digest']
    job_parameters = params['jobParameters'];

    # Make sure the image exists on the server before we go further
    _ensure_image_on_server(task, repository, tag, digest, container)

    output_file = 'description.json'

    run_command = '%s run $IMAGE_NAME' % container
    # Shifter has a pretty different sytax so special case it.
    if container == 'shifter':
        run_command = 'shifter --image=$IMAGE_NAME --entrypoint --'

    image_name = '%s:%s' % (repository, tag)
    if container == 'singularity':
        image_name = digest_to_sif(digest)

    commands = setup_commands + [
        'IMAGE_NAME=%s' % image_name,
        '%s -d > %s' % (run_command, output_file)
    ]

    body = {
        # ensure there are no special characters in the submission script name
        'name': 'desc_%s' % re.sub('[^a-zA-Z0-9]', '_', repository),
        'commands': commands,
        'input': [
            {
              'folderId': description_folder['_id'],
              'path': '.'
            }
        ],
        'output': [
            {
              'folderId': description_folder['_id'],
              'path': '.'
            }
        ],
        'uploadOutput': False,
        'params': job_parameters
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))

    return job

@cumulus.taskflow.task
def postprocess_description(task, _, input_, user, cluster, image, run_parameters, root_folder, description_job, description_folder):
    task.taskflow.logger.info('Processing the output of the container description job.')

    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    # Refresh state of job
    description_job = client.get('jobs/%s' % description_job['_id'])

    upload_job_output_to_folder(cluster, description_job, girder_token=task.taskflow.girder_token)

    description_items = list(client.listItem(description_folder['_id']))

    description_file = None
    for item in description_items:
        if item['name'] == 'description.json':
            files = list(client.listFile(item['_id']))
            if len(files) != 1:
                log_std_err(task, client, description_folder)
                log_and_raise(task, 'Expecting a single file under item, found: %s' % len(files))
            description_file = files[0]

    if description_file is None:
        log_std_err(task, client, description_folder)
        log_and_raise(task, 'The container does not implement correctly the --description flag')

    with client.session() as session:
        # If we have a NEWT session id we need set as a cookie so the redirect
        # to the NEWT API works ( is authenticated ).
        newt_session_id = parse('newt.sessionId').find(user)
        if newt_session_id:
            newt_session_id = newt_session_id[0].value
            session.cookies.set('newt_sessionid', newt_session_id)

        with tempfile.TemporaryFile() as tf:
            client.downloadFile(description_file['_id'], tf)
            tf.seek(0)
            container_description = json.loads(tf.read().decode())

    # Add code name and version to the taskflow metadata
    code = {
        'name': container_description.get('name'),
        'version': container_description.get('version')
    }
    task.taskflow.set_metadata('code', code)

    # remove temporary description folder
    client.delete('folder/%s' % description_folder['_id'])

    setup_input.delay(input_, cluster, image, run_parameters, root_folder, container_description)

@cumulus.taskflow.task
def setup_input(task, input_, cluster, image, run_parameters, root_folder, container_description):
    task.taskflow.logger.info('Setting up the calculation input files.')

    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    if cluster.get('name') == 'cori':
        cluster = get_cori(client)

    if '_id' not in cluster:
        log_and_raise(task, 'Invalid cluster configurations: %s' % cluster)

    calculation_ids = parse('calculations').find(input_)
    if not calculation_ids:
        log_and_raise(task, 'Unable to extract calculation ids.')

    calculation_ids = calculation_ids[0].value
    calculations = [client.get('calculations/%s' % x) for x in calculation_ids]
    molecule_ids = [x['moleculeId'] for x in calculations]
    geometry_ids = [x.get('geometryId') for x in calculations]

    input_parameters = [x.get('input', {}).get('parameters', {})
                        for x in calculations]

    # For now, we only allow multiple calculations if all of the input
    # parameters are the same. Raise an error if they differ.
    if not all(x == input_parameters[0] for x in input_parameters):
        msg = ('For running multiple calculations, all input parameters must '
               'currently be identical')
        log_and_raise(task, msg)

    input_parameters = input_parameters[0]

    input_format = container_description['input']['format']

    # Fetch the starting geometries
    geometry_data = []
    for molecule_id, geometry_id in zip(molecule_ids, geometry_ids):

        path = 'molecules/%s/' % molecule_id
        if geometry_id:
            # It is preferred to use the geometry if we have it
            path += 'geometries/%s/' % geometry_id

        path += '%s' % input_format

        r = client.get(path, jsonResp=False)

        if r.status_code != 200:
            raise Exception('Failed to get molecule in format: ' + input_format)

        if input_format == 'cjson':
            geometry_data.append(json.dumps(r.json()))
        else:
            geometry_data.append(r.content.decode('utf-8'))

    # The folder where the input geometry and input parameters are
    input_folder = client.createFolder(root_folder['_id'], 'input')
    # The folder where the converted output will be at the end of the job
    output_folder = client.createFolder(root_folder['_id'], 'output')
    # The folder where the raw input/output files of the specific code are stored
    scratch_folder = client.createFolder(root_folder['_id'], 'scratch')
    # The folder where the cluster stdout and stderr is saved
    run_folder = client.createFolder(root_folder['_id'], 'run')

    # Save the input parameters to file
    with tempfile.TemporaryFile() as fp:
        fp.write(json.dumps(input_parameters).encode())
        # Get the size of the file
        size = fp.seek(0, 2)
        fp.seek(0)
        name = 'input_parameters.json'
        input_parameters_file = client.uploadFile(input_folder['_id'],  fp,
                                                  name, size,
                                                  parentType='folder')

    # Save the input geometries to files
    for i, data in enumerate(geometry_data):
        with tempfile.TemporaryFile() as fp:
            fp.write(data.encode())
            # Get the size of the file
            size = fp.seek(0, 2)
            fp.seek(0)
            name = 'geometry_' + str(i + 1) + '.%s' % input_format
            input_geometry_file = client.uploadFile(input_folder['_id'], fp,
                                                    name, size,
                                                    parentType='folder')

    submit_calculation.delay(input_, cluster, image, run_parameters, root_folder, container_description, input_folder, output_folder, scratch_folder, run_folder)

def _create_job(task, input_, cluster, image, run_parameters, container_description, input_folder, output_folder, scratch_folder, run_folder):
    params = _get_job_parameters(task, cluster, image, run_parameters)
    container = params['container']
    repository = params['repository']
    digest = params['digest']
    host_dir = params['hostDir']
    guest_dir = params['guestDir']
    job_dir = params['jobDir']
    job_parameters = params['jobParameters'];

    task.taskflow.logger.info('Creating %s job' % repository)

    input_format = container_description['input']['format']
    output_format = container_description['output']['format']

    input_dir = os.path.join(guest_dir, job_dir, 'input')
    output_dir = os.path.join(guest_dir, job_dir, 'output')
    scratch_dir = os.path.join(guest_dir, job_dir, 'scratch')

    parameters_filename = os.path.join(input_dir, 'input_parameters.json')

    calculation_ids = parse('calculations').find(input_)[0].value

    geometry_filenames = []
    output_filenames = []
    for i in range(len(calculation_ids)):
        geometry_filenames.append(os.path.join(input_dir, 'geometry_' + str(i + 1) + '.%s' % input_format))
        output_filenames.append(os.path.join(output_dir, 'output_' + str(i + 1) + '.%s' % output_format))

    output = [
        {
            'folderId': output_folder['_id'],
            'path': './output'
        },
        {
            'folderId': run_folder['_id'],
            'path': '.',
            'exclude': [
                'input',
                'output',
                'scratch'
            ]
        }
    ]

    keep_scratch = run_parameters.get('keepScratch', False)
    if keep_scratch:
        output.append({
            'folderId': scratch_folder['_id'],
            'path': './scratch'
        })

    commands = [
        'mkdir output',
        'mkdir scratch'
    ]

    # Each contain has a different bind arg
    bind_args = {
        'docker': '-v',
        'singularity': '-B',
        'shifter': '-V'
    }

    mount_option = '%s %s:%s' % (bind_args[container], host_dir, guest_dir)

    container_args = '-p %s -s %s' % (
        parameters_filename, scratch_dir
    )

    for g, o in zip(geometry_filenames, output_filenames):
        container_args += ' -g %s -o %s' % (g, o)

    image_str = image.get('repository') + ':' + image.get('tag')

    if container == 'singularity':
        # Include the path to the singularity dir, and the extension
        image_str = digest_to_sif(digest)

    if container != 'shifter':
        commands.append('%s run %s %s %s' % (
            container, mount_option, image_str, container_args
        ))
    # Shifters syntax is pretty different so special case it
    else:
        commands.append('shifter %s --image=%s --entrypoint -- %s'  % (
            mount_option, image_str, container_args
        ))

    if is_nersc(cluster):
        # NERSC specific options
        job_parameters.update({
            'numberOfNodes': 1,
            'queue': 'debug',
            'constraint': 'haswell',
            'account': os.environ.get('OC_ACCOUNT')
        })

    body = {
        # ensure there are no special characters in the submission script name
        'name': 'run_%s' % re.sub('[^a-zA-Z0-9]', '_', repository),
        'commands': commands,
        'input': [
            {
              'folderId': input_folder['_id'],
              'path': './input'
            }
        ],
        'output': output,
        'uploadOutput': False,
        'params': job_parameters
    }

    client = create_girder_client(
                task.taskflow.girder_api_url, task.taskflow.girder_token)

    job = client.post('jobs', data=json.dumps(body))
    task.taskflow.set_metadata('jobs', [job])

    return job

@cumulus.taskflow.task
def submit_calculation(task, input_, cluster, image, run_parameters, root_folder, container_description, input_folder, output_folder, scratch_folder, run_folder):
    job = _create_job(task, input_, cluster, image, run_parameters, container_description, input_folder, output_folder, scratch_folder, run_folder)

    girder_token = task.taskflow.girder_token
    task.taskflow.set_metadata('cluster', cluster)

    # Now download and submit job to the cluster
    task.taskflow.logger.info('Uploading the input files to the cluster.')
    download_job_input_folders(cluster, job,
                               girder_token=girder_token, submit=False)

    task.taskflow.logger.info('Submitting job %s to the queue.' % job['_id'])

    submit_job(cluster, job, girder_token=girder_token, monitor=False)

    task.taskflow.logger.info('Submitted job %s to cluster.' % job['_id'])

    monitor_job.apply_async((cluster, job), {'girder_token': girder_token,
                                             'monitor_interval': 10},
                                             countdown=countdown(cluster),
                            link=postprocess_job.s(input_, cluster, image, run_parameters, root_folder, container_description, input_folder, output_folder, scratch_folder, run_folder, job))

@cumulus.taskflow.task
def postprocess_job(task, _, input_, cluster, image, run_parameters, root_folder, container_description, input_folder, output_folder, scratch_folder, run_folder, job):
    task.taskflow.logger.info('Processing the results of the calculation.')
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)

    # Refresh state of job
    job = client.get('jobs/%s' % job['_id'])

    upload_job_output_to_folder(cluster, job, girder_token=task.taskflow.girder_token)

    # remove temporary input folder, this data is attached to the calculation model
    client.delete('folder/%s' % input_folder['_id'])

    # clean up the scratch folder
    keep_scratch = run_parameters.get('keepScratch', False)
    if keep_scratch:
        scratch_folder_id = scratch_folder['_id']
    else:
        client.delete('folder/%s' % scratch_folder['_id'])
        scratch_folder_id = None

    # ingest the output of the calculation
    output_format = container_description['output']['format']
    output_files = []
    output_items = list(client.listItem(output_folder['_id']))
    for i in range(len(input_['calculations'])):
        output_file = None
        for item in output_items:
            if item['name'] == 'output_' + str(i + 1) + '.%s' % output_format:
                files = list(client.listFile(item['_id']))
                if len(files) != 1:
                    log_std_err(task, client, run_folder)
                    log_and_raise(task, 'Expecting a single file under item, found: %s' % len(files))
                output_file = files[0]
                break

        if output_file is None:
            # Log the job stderr
            log_std_err(task, client, run_folder)
            log_and_raise(task, 'The calculation did not produce any output file.')

        output_files.append(output_file)

    # remove the run folder, only useful to access the stdout and stderr after the job is done
    client.delete('folder/%s' % run_folder['_id'])

    # Now call endpoint to ingest result
    params = {
        'detectBonds': True
    }

    task.taskflow.logger.info('Uploading the results of the calculation to the database.')

    code = task.taskflow.get_metadata('code')
    if isinstance(code, dict):
        # Get the contents of "code" to set it below
        code = code.get('code')

    for i, output_file in enumerate(output_files):
        body = {
            'fileId': output_file['_id'],
            'format': output_format,
            'public': True,
            'image': image, # image now also has a digest field, add it to the calculation
            'scratchFolderId': scratch_folder_id,
            'code': code
        }

        client.put('calculations/%s' % input_['calculations'][i], parameters=params, json=body)

    task.taskflow.logger.log(STATUS_LEVEL, 'Done!')


def _ensure_image_on_server(task, repository, tag, digest, container='docker'):
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)
    params = {
      'repository': repository,
      'tag': tag,
      'digest': digest
    }
    r = client.get('images', params)
    images = r['results']
    if len(images) < 1:
        log_and_raise(task, 'Image not found on the server.')

    if container not in images[0]:
        msg = 'Image does not have container type: ' + container
        log_and_raise(task, msg)


def _get_digest(task, repository, tag):
    client = create_girder_client(
        task.taskflow.girder_api_url, task.taskflow.girder_token)
    params = {
      'repository': repository,
      'tag': tag
    }
    r = client.get('images', params)
    images = r['results']
    if len(images) < 1:
        log_and_raise(task, 'Image not found on the server.')

    # The digest should definitely be here
    return images[0]['digest']



import os

from ._girder import GirderClient
from ._singleton import Singleton
from ._utils import lookup_file

@Singleton
class JupyterHub(object):

    def __init__(self):
        url = os.environ.get('OC_JUPYTERHUB_URL')
        self.url = url
        self.file = None
        
        if self.url is not None and GirderClient().client is not None:
            self.file = lookup_file(GirderClient().client, self.url)



from ._utils import (
    hash_object, camel_to_space, parse_image_name, calculate_rmsd
)
from .io import (
    CjsonReader, Cp2kReader, NWChemJsonReader, OrcaReader, Psi4Reader
)
from .api import (
    load, find_structure, find_calculation, find_molecule, monitor, queue,
    find_spectra, import_structure, run_calculations
)



import os
import inspect
import json
from jsonpath_rw import parse
import urllib
import urllib.parse

from girder_client import HttpError

from ._girder import GirderClient
from ._jupyterhub import JupyterHub
from ._cluster import Cluster
from ._molecule import Molecule
from ._data import MoleculeProvider, CalculationProvider
from ._utils import (
    fetch_or_create_queue, hash_object, parse_image_name, mol_has_3d_coords,
    get_oc_token_obj
)

class GirderMolecule(Molecule):
    '''
    Derived version that allows calculations to be initiated on using Girder
    '''
    def __init__(self, _id, cjson=None):
        if cjson is None:
            try:
                cjson = GirderClient().get('molecules/%s/cjson' % _id)
            except HttpError as ex:
                if ex.status == 404:
                    # The molecule does not have 3D coordinates
                    cjson = None
                else:
                    raise

        super(GirderMolecule, self).__init__(MoleculeProvider(cjson, _id))
        self._id = _id

    def calculate(self, image_name, input_parameters, geometry_id=None, run_parameters=None, force=False):
        molecule_id = self._id
        try:
            calculations = _fetch_or_submit_calculations([molecule_id],
                                                         image_name,
                                                         input_parameters,
                                                         [geometry_id],
                                                         run_parameters, force)
        except Exception as e:
            image_not_found_msg = 'Image not found on the server'
            if e.args and image_not_found_msg in e.args[0]:
                print(image_not_found_msg)
                return

            raise

        if calculations:
            return _calculation_result(calculations[0], molecule_id)

    def energy(self, image_name, input_parameters, geometry_id=None, run_parameters=None, force=False):
        params = {'task': 'energy'}
        params.update(input_parameters)
        return self.calculate(image_name, params, geometry_id, run_parameters, force)

    def optimize(self, image_name, input_parameters, geometry_id=None, run_parameters=None, force=False):
        params = {'task': 'optimize'}
        params.update(input_parameters)
        return self.calculate(image_name, params, geometry_id, run_parameters, force)

    def frequencies(self, image_name, input_parameters, geometry_id=None, run_parameters=None, force=False):
        params = {'task': 'frequencies'}
        params.update(input_parameters)
        return self.calculate(image_name, params, geometry_id, run_parameters, force)

    def set_name(self, name):
        body = {
            'name': name
        }
        GirderClient().patch('molecules/%s' % self._id, json=body)

    def add_geometry(self, cjson):
        GirderClient().post('molecules/%s/geometries' % self._id, json=cjson)

class CalculationResult(Molecule):

    def __init__(self, _id=None, properties=None, molecule_id=None):
        super(CalculationResult, self).__init__(CalculationProvider(_id, molecule_id))
        self._id = _id
        self._properties = properties
        self._molecule_id = molecule_id
        self._optimized_geometry_id = None

    def data(self):
        return self._provider.cjson

    @property
    def optimized_geometry_id(self):
        if not self._optimized_geometry_id:
            # Try to get it...
            result = GirderClient().get('calculations/%s' % self._id)
            if 'optimizedGeometryId' in result:
                self._optimized_geometry_id = result['optimizedGeometryId']
            else:
                print('None')

        return self._optimized_geometry_id

    @property
    def frequencies(self):
        import warnings
        warnings.warn("Use the 'vibrations' property to display normal modes")
        return self.vibrations

    def delete(self):
        return _delete_calculation(self._id)

class AttributeInterceptor(object):
    def __init__(self, wrapped, value, intercept_func=lambda : True):
        self._wrapped = wrapped
        self._value = value
        self._intercept_func = intercept_func

    def unwrap(self):
        return self._wrapped

    def __getattr__(self, name):
        # Use object's implementation to get attributes, otherwise
        # we will get recursion
        _wrapped = object.__getattribute__(self, '_wrapped')
        _value = object.__getattribute__(self, '_value')
        _intercept_func = object.__getattribute__(self, '_intercept_func')

        attr = object.__getattribute__(_wrapped, name)
        if _intercept_func():
            if inspect.ismethod(attr):
                def pending(*args, **kwargs):
                    return _value
                return pending
            else:
                return AttributeInterceptor(attr, _value, _intercept_func)
        else:
            return attr

class PendingCalculationResultWrapper(AttributeInterceptor):
    def __init__(self, calculation, taskflow_id=None):
        try:
            from ._notebook import CalculationMonitor
            if taskflow_id is None:
                taskflow_id = calculation._properties['taskFlowId']

            table = CalculationMonitor({
                'taskFlowIds': [taskflow_id],
                'girderToken': GirderClient().token,
                'girderApiUrl': GirderClient().url
            })
        except ImportError:
            # Outside notebook just print message
            table = 'Pending calculations .... '

        # Only intercept when the taskflow is not complete
        def intercept():
            return _fetch_taskflow_status(taskflow_id) != 'complete'

        super(PendingCalculationResultWrapper, self).__init__(calculation,
                                                              table, intercept)

def _fetch_calculation(molecule_id, image_name, input_parameters, geometry_id=None):
    repository, tag = parse_image_name(image_name)
    input_params_quoted = urllib.parse.quote(json.dumps(input_parameters))
    parameters = {
        'moleculeId': molecule_id,
        'inputParameters': input_params_quoted,
        'imageName': '%s:%s' % (repository, tag)
    }

    if geometry_id:
        parameters['geometryId'] = geometry_id

    res = GirderClient().get('calculations', parameters)
    if 'results' not in res or len(res['results']) < 1:
        return None

    return res['results'][0]

def _nersc():
    oc_token_obj = get_oc_token_obj()
    site = oc_token_obj.get('site')
    site = os.environ.get('OC_SITE', site)
    return site == 'NERSC'

def _submit_calculations(cluster_id, pending_calculation_ids, image_name,
                         run_parameters):
    if run_parameters is None:
        run_parameters = {}

    repository, tag = parse_image_name(image_name)

    body = {}
    body['taskFlowBody'] = {
        'taskFlowClass': 'taskflows.OpenChemistryTaskFlow',
        'meta': {
            'calculationIds': pending_calculation_ids,
            'image': {
                'repository': repository,
                'tag': tag
            }
        }
    }

    body['taskFlowInput'] = {
        'input': {
            'calculations': pending_calculation_ids
        },
        'image': {
            'repository': repository,
            'tag': tag
        },
        'runParameters': run_parameters
    }

    if cluster_id:
        body['taskFlowInput']['cluster'] = {'_id': cluster_id}

    # This returns the taskflow id
    return GirderClient().post('launch_taskflow/launch', json=body)

def _fetch_taskflow_status(taskflow_id):
    r = GirderClient().get('taskflows/%s/status' % taskflow_id)

    return r['status']

def _ensure_image_on_server(repository, tag, container='docker', digest=None):
    params = {
      'repository': repository,
      'tag': tag
    }

    if digest is not None:
        params['digest'] = digest

    r = GirderClient().get('images', params)
    images = r['results']
    if len(images) < 1:
        raise Exception('Image not found on the server')

    if container not in images[0]:
        raise Exception('Container type not found in image')

def _create_pending_calculation(molecule_id, image_name, input_parameters, geometry_id=None):
    repository, tag = parse_image_name(image_name)

    # Verify that the image is on the server before going any further
    _ensure_image_on_server(repository, tag)

    notebooks = []
    if JupyterHub().file is not None:
        notebooks.append(JupyterHub().file['_id'])

    body = {
        'moleculeId': molecule_id,
        'cjson': None,
        'public': True,
        'properties': {
            'pending': True
        },
        'input': {
            'parameters': input_parameters,
        },
        'image': {
            'repository': repository,
            'tag': tag
        },
        'notebooks': notebooks
    }

    if geometry_id is not None:
        body['geometryId'] = geometry_id

    calculation = GirderClient().post('calculations', json=body)

    return calculation

def _delete_calculation(calculation_id):
    GirderClient().delete('calculations/%s' % calculation_id)

def _fetch_or_submit_calculations(molecule_ids, image_name, input_parameters,
                                  geometry_ids=None, run_parameters=None,
                                  force=False):

    try:
        _check_required_coords(molecule_ids, image_name)
    except Exception as e:
        print(str(e))
        return []

    if geometry_ids is None:
        geometry_ids = [None] * len(molecule_ids)

    calculations = []
    pending_calculations = []
    for molecule_id, geometry_id in zip(molecule_ids, geometry_ids):
        calculation = _fetch_calculation(molecule_id, image_name,
                                         input_parameters, geometry_id)
        if calculation is None or force:
            calculation = _create_pending_calculation(molecule_id, image_name,
                                                      input_parameters,
                                                      geometry_id)
            pending_calculations.append(calculation)
        else:
            # If we already have a calculation tag it with this notebooks id
            notebooks = calculation.setdefault('notebooks', [])
            if JupyterHub().file is not None:
                notebooks.append(JupyterHub().file['_id'])

            body = {
                'notebooks': notebooks
            }
            GirderClient().patch('calculations/%s/notebooks' %
                                 calculation['_id'],
                                 json=body)

        calculations.append(calculation)

    if len(pending_calculations) != 0:
        calc_ids = [x['_id'] for x in pending_calculations]
        taskflow_id = _submit_calculations(Cluster().id, calc_ids,
                                           image_name, run_parameters)

        for i, calculation in enumerate(pending_calculations):
            # Patch calculation to include taskflow id
            props = calculation['properties']
            props['taskFlowId'] = taskflow_id
            pending_calculations[i] = GirderClient().put(
                'calculations/%s/properties' % calculation['_id'], json=props)
            index = calculations.index(calculation)
            calculations[index] = pending_calculations[i]

    return calculations

def _calculation_result(calculation, molecule_id):
    pending = parse('properties.pending').find(calculation)
    if pending:
        pending = pending[0].value

    result = CalculationResult(calculation['_id'], calculation['properties'],
                               molecule_id)

    if pending:
        taskflow_id = parse('properties.taskFlowId').find(calculation)
        if taskflow_id:
            taskflow_id = taskflow_id[0].value
        else:
            taskflow_id = None

        result = PendingCalculationResultWrapper(result, taskflow_id)

    return result

def _3d_coords_required(image_name):
    # We currently have a whitelist of programs that do not require
    # 3D coordinates. Assume 3D coords are required for any other
    # program.
    # TODO: make some way for us to be able to get information about
    # the images here, so we can do this via the description.json file.
    white_list = [
        'chemml'
    ]

    for program in white_list:
        if program in image_name:
            return False

    return True

def _mol_has_3d_coords(mol_id):
    mol = GirderClient().get('molecules/%s' % mol_id)
    return mol_has_3d_coords(mol)

def _generate_3d_coords(mol_id):
    GirderClient().post('molecules/%s/3d' % mol_id)

def _check_required_coords(mol_ids, image_name):
    raise_exception = False
    if _3d_coords_required(image_name):
        for mol_id in mol_ids:
            if not _mol_has_3d_coords(mol_id):
                _generate_3d_coords(mol_id)
                raise_exception = True

    if raise_exception:
        msg = 'Generating 3D coordinates, please re-run the calculation soon'
        raise Exception(msg)
